/*****************************************************************************
 * This file is part of gfxprim library.                                     *
 *                                                                           *
 * Gfxprim is free software; you can redistribute it and/or                  *
 * modify it under the terms of the GNU Lesser General Public                *
 * License as published by the Free Software Foundation; either              *
 * version 2.1 of the License, or (at your option) any later version.        *
 *                                                                           *
 * Gfxprim is distributed in the hope that it will be useful,                *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Lesser General Public License for more details.                           *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public          *
 * License along with gfxprim; if not, write to the Free Software            *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,                        *
 * Boston, MA  02110-1301  USA                                               *
 *                                                                           *
 * Copyright (C) 2009-2013 Cyril Hrubis <metan@ucw.cz>                       *
 *                                                                           *
 *****************************************************************************/

#include <string.h>
#include <errno.h>

#include "core/GP_Debug.h"

#include "GP_Backends.h"
#include "GP_BackendInit.h"

static char *next_param(char *params)
{
	for (;;) {
		switch (*params) {
		case ':':
			*params = '\0';
			return params + 1;
		break;
		case '\0':
			return NULL;
		break;
		}
		params++;
	}
}

static int parse_x11_params(char *params, GP_Size *w, GP_Size *h,
                            enum GP_BackendX11Flags *flags)
{
	char *param;

	if (!params)
		return 0;

	do {
		param = params;
		params = next_param(params);

		if (!strcasecmp(param, "use_root")) {
			*flags |= GP_X11_USE_ROOT_WIN;
			GP_DEBUG(1, "X11: Using root window");
			continue;
		}

		if (!strcasecmp(param, "create_root")) {
			*flags |= GP_X11_CREATE_ROOT_WIN;
			GP_DEBUG(1, "X11: Creating root window");
			continue;
		}

		if (!strcasecmp(param, "disable_shm")) {
			*flags |= GP_X11_DISABLE_SHM;
			GP_DEBUG(1, "X11: Disabling SHM");
			continue;
		}

		if (!strcasecmp(param, "fs")) {
			*flags |= GP_X11_FULLSCREEN;
			GP_DEBUG(1, "X11: Enabling fullscreen");
			continue;
		}

		/*
		 * Accepts only string with format "intxint" or "intXint"
		 */
		int sw, sh;
		unsigned int n;

		if (sscanf(param, "%i%*[xX]%i%n", &sw, &sh, &n) == 2 && n == strlen(param)) {
			*w = sw;
			*h = sh;
			continue;
		}

		GP_WARN("X11: Invalid parameters '%s'", param);
		errno = EINVAL;
		return 1;
	} while (params);

	return 0;
}

static GP_Backend *x11_init(char *params, const char *caption)
{
	GP_Size w = 640, h = 480;
	enum GP_BackendX11Flags flags = 0;

	if (parse_x11_params(params, &w, &h, &flags))
		return NULL;

	return GP_BackendX11Init(NULL, 0, 0, w, h, caption, flags);
}

static int parse_sdl_params(char *params, GP_Size *w, GP_Size *h,
                            GP_Size *bpp, uint8_t *flags)
{
	char *param;

	if (!params)
		return 0;

	do {
		param = params;
		params = next_param(params);

		if (!strcasecmp(param, "FS")) {
			*flags |= GP_SDL_FULLSCREEN;
			GP_DEBUG(1, "SDL fullscreen enabled");
			continue;
		}

		if (!strcmp(param, "8")) {
			*bpp = 8;
			GP_DEBUG(1, "SDL depth set to 8");
			continue;
		}

		if (!strcmp(param, "16")) {
			*bpp = 16;
			GP_DEBUG(1, "SDL depth set to 16");
			continue;
		}

		if (!strcmp(param, "24")) {
			*bpp = 24;
			GP_DEBUG(1, "SDL depth set to 24");
			continue;
		}

		if (!strcmp(param, "32")) {
			*bpp = 32;
			GP_DEBUG(1, "SDL depth set to 32");
			continue;
		}

		/*
		 * Accepts only string with format "intxint" or "intXint"
		 */
		int sw, sh;
		unsigned int n;

		if (sscanf(param, "%i%*[xX]%i%n", &sw, &sh, &n) == 2 && n == strlen(param)) {
			*w = sw;
			*h = sh;
			continue;
		}

		GP_WARN("SDL: Invalid parameters '%s'", param);
		errno = EINVAL;
		return 1;
	} while (params);

	return 0;
}

static GP_Backend *sdl_init(char *params, const char *caption)
{
	GP_Size w = 0, h = 0, bpp = 0;
	uint8_t flags = GP_SDL_RESIZABLE;

	if (parse_sdl_params(params, &w, &h, &bpp, &flags))
		return NULL;

	return GP_BackendSDLInit(w, h, bpp, flags, caption);
}

static int parse_fb_params(char *params, int *flags, const char **fb)
{
	char *param;

	if (!params)
		return 0;

	do {
		param = params;
		params = next_param(params);

		if (!strcasecmp(param, "no_shadow")) {
			*flags &= ~GP_FB_SHADOW;
			GP_DEBUG(1, "Shadow framebuffer disabled");
			continue;
		}

		if (!strcasecmp(param, "new_console")) {
			*flags |= GP_FB_ALLOC_CON;
			GP_DEBUG(1, "Console allocation enabled");
			continue;
		}

		*fb = param;

		if (strncmp(*fb, "/dev/", 5))
			GP_WARN("Console dev set to '%s', are you sure?", *fb);

		GP_DEBUG(1, "Framebuffer console set to '%s'", *fb);

	} while (params);

	return 0;
}

static GP_Backend *fb_init(char *params, const char *caption)
{
	const char *fb = "/dev/fb0";

	(void) caption;

	int flags = GP_FB_INPUT_KBD | GP_FB_SHADOW;

	parse_fb_params(params, &flags, &fb);

	return GP_BackendLinuxFBInit(fb, flags);
}

static GP_Backend *aa_init(char *params, const char *caption)
{
	(void) caption;
	(void) params;

	return GP_BackendAALibInit();
}

struct backend_init {
	const char *name;
	GP_Backend *(*init)(char *params, const char *caption);
	const char *usage;
	const char *help[10];
};

static GP_Backend *do_help(char *params, const char *caption);

static struct backend_init backends[] = {
	{.name  = "X11",
	 .init  = x11_init,
	 .usage = "X11:[WxH]:[use_root]:[create_root]:[disable_shm]",
	 .help  = {"use_root    - starts the backend in the root window",
	           "              (w and h, if set, are ignored)",
	           "create_root - starts the backend in newly created",
	           "              root window (w and h, if set, are ignored)",
	           "disable_shm - disable MIT SHM even if available",
	           "fs          - start fullscreen",
	           NULL}
	},
	{.name  = "SDL",
	 .init  = sdl_init,
	 .usage = "SDL:[fs]:[8]:[16]:[24]:[32]:[WxH]",
	 .help  = {"fs  - Full Screen mode",
	           "8   - Sets 8bpp",
	           "16  - Sets 16bpp",
	           "24  - Sets 24bpp",
	           "32  - Sets 32bpp",
	           "WxH - Display Size",
	           NULL}
	},
	{.name  = "FB",
	 .init  = fb_init,
	 .usage = "fb:[no_shadow]:[new_console]:[/dev/fbX]",
	 .help  = {"no_shadow   - turns off shadow buffer",
	           "new_console - allocate new console",
	           NULL}
	},
	{.name  = "AA",
	 .init  = aa_init,
	 .usage = "AA",
	 .help  = {NULL}
	},
	{.name = "help",
	 .init = do_help
	},
	{.name = NULL}
};

static GP_Backend *do_help(char *params, const char *caption)
{
	struct backend_init *i;
	unsigned int j;

	(void) params;
	(void) caption;

	for (i = backends; (i+1)->name; i++) {
		fprintf(stderr, "Backend %s\n\n %s\n\n",
		        i->name, i->usage);
		if (i->help[0]) {
			for (j = 0; i->help[j]; j++)
				fprintf(stderr, " %s\n", i->help[j]);
			fprintf(stderr, "\n");
		}
	}

	return NULL;
}

static struct backend_init *get_backend(const char *name)
{
	struct backend_init *i;

	for (i = backends; i->name; i++) {
		if (!strcasecmp(name, i->name))
			return i;
	}

	return NULL;
}

static GP_Backend *init_backend(const char *name, char *params,
                                const char *caption)
{
	struct backend_init *init = get_backend(name);
	GP_Backend *ret;

	if (!init) {
		GP_WARN("Invalid backend name '%s'", name);
		errno = EINVAL;
		return NULL;
	}

	ret = init->init(params, caption);

	return ret;
}

GP_Backend *GP_BackendInit(const char *params, const char *caption)
{
	if (params == NULL) {
		do_help(NULL, NULL);
		return NULL;
	}

	/* parse backend name */
	int i, len = strlen(params);
	char buf[len+1], *backend_params = NULL;

	strcpy(buf, params);

	for (i = 0; i < len; i++) {
		if (buf[i] == ':') {
			buf[i] = '\0';
			backend_params = buf + i + 1;
			break;
		}
	}

	GP_DEBUG(1, "Have backend name '%s' params '%s'", buf, backend_params);

	return init_backend(buf, backend_params, caption);
}
